/*
 * Curses UI for slide puzzle game
 * Copyright © 2022 Nick Bowler
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <config.h>
#include <stdlib.h>
#include <locale.h>
#include <assert.h>
#include <getopt.h>
#include <curses.h>

#include "help.h"
#include "version.h"
#include "cursesopt.h"
#include "game.h"

#define MIN(a, b) ((a) < (b) ? (a) : (b))
#define MAX(a, b) ((a) > (b) ? (a) : (b))

static const char *progname = "rrace";
static const struct option lopts[] = { LOPTS_INITIALIZER, {0} };

enum {
	WINDOW_TILEBORDER,
	WINDOW_TILEFILL,
	WINDOW_AREA,
	WINDOW_MAX,
};

static struct app_state {
	struct board board;

	WINDOW *gamewin[WINDOW_MAX], *goalwin[WINDOW_MAX];
	int cursor;
} state;

static void print_version(void)
{
	version_print_head("rrace-curses", stdout);
	puts("License GPLv3+: GNU GPL version 3 or any later version");
	puts("This is free software: you are free to change and redistribute it.");
	puts("There is NO WARRANTY, to the extent permitted by law.");
}

static void print_usage(FILE *f)
{
	fprintf(f, "Usage: %s [options]\n", progname);
	if (f != stdout)
		fprintf(f, "Try %s --help for more information.\n", progname);
}

static void print_help(void)
{
	struct lopt_help help = {0};
	const struct option *opt;

	print_usage(stdout);

	putchar('\n');
	puts("Options:");
	for (opt = lopts; opt->name; opt++) {
		if (!lopt_get_help(opt, &help))
			continue;
		help_print_option(opt, help.arg, help.desc, 20);
	}
	putchar('\n');

	printf("Report bugs to <%s>.\n", PACKAGE_BUGREPORT);
}

static void draw_tile(WINDOW **win, unsigned colour, unsigned selected,
                      unsigned x, unsigned y, unsigned start_column)
{
	WINDOW *border = win[WINDOW_TILEBORDER], *fill = win[WINDOW_TILEFILL];
	int w, h, attr, ch, bc = selected ? '#' : 0;

	assert(colour < TILE_MAX);
	attr = COLOR_PAIR(colour);
	switch (colour) {
	case TILE_RED:    ch = 'X'; break;
	case TILE_ORANGE: ch = '|'; break;
	case TILE_GREEN:  ch = '+'; break;
	case TILE_YELLOW: ch = '~'; attr |= A_BOLD; break;
	case TILE_BLUE:   ch = 'o'; attr |= A_BOLD; break;
	case TILE_WHITE:  ch = '.'; attr |= A_BOLD; break;

	case TILE_EMPTY: attr = A_BOLD|COLOR_PAIR(TILE_MAX);
	}

	getmaxyx(border, h, w);
	w = 2*(w+1)/2;

	if (mvwin(border, 2+h*y, start_column+w*x) == ERR)
		return;

	if (colour != TILE_EMPTY) {
		mvderwin(fill, 1, 1);
		wbkgdset(fill, A_REVERSE|attr|ch);
		werase(fill);
	} else {
		werase(border);
	}

	if (bc || colour) {
		wattrset(border, attr);
		wborder(border, bc, bc, bc, bc, bc, bc, bc, bc);
	}

	wnoutrefresh(border);
}

static int
redraw_tile(WINDOW **win, unsigned x, unsigned y, unsigned start_column,
	    uint_fast32_t bit0, uint_fast32_t bit1, uint_fast32_t bit2,
	    unsigned selected)
{
	uint_fast32_t pos = board_position(x, y);
	unsigned char tile = 0;

	if (bit0 & pos) tile |= 1;
	if (bit1 & pos) tile |= 2;
	if (bit2 & pos) tile |= 4;
	assert(tile < TILE_MAX);

	draw_tile(win, tile, selected, x, y, start_column);
	return tile;
}

static void redraw_area_border(WINDOW **win, unsigned x, unsigned sz)
{
	int w, h, tr = 0, rs = 0, br = 0, bs = 0, bl = 0;
	WINDOW *area = win[WINDOW_AREA];

	getmaxyx(stdscr, h, w);

	if (h <= 2) {
		bl = ACS_ULCORNER, br = ACS_URCORNER;
	} else if (h <= 3*sz+2) {
		bl = br = ACS_VLINE, bs = ' ';
	}

	if (w <= x+1) {
		tr = ACS_ULCORNER, br = ACS_LLCORNER;
	} else if (w <= 6*sz+x+2) {
		tr = ACS_HLINE, rs = ' ';
		br = br ? ' ' : ACS_HLINE;
	}

	wborder(area, 0, rs, 0, bs, 0, tr, bl, br);
	wnoutrefresh(area);
}

static void curs_redraw_game(struct app_state *state, uint_fast32_t mask)
{
	uint_least32_t *gp = state->board.game;
	int i;

	if (mask == -1)
		redraw_area_border(state->gamewin, 2, 5);

	for (i = 0; i < 25; i++) {
		if (mask & 1) {
			redraw_tile(state->gamewin, i%5, i/5,
			            4, gp[0], gp[1], gp[2],
				    i == state->cursor);
		}
		mask >>= 1;
	}
}

static void curs_redraw_goal(struct app_state *state, uint_fast32_t mask)
{
	uint_least16_t *gp = state->board.goal;
	int i, x, y;

	if (!state->goalwin[WINDOW_AREA])
		return;

	getbegyx(state->goalwin[WINDOW_AREA], y, x);
	if (mask == -1)
		redraw_area_border(state->goalwin, x, 3);

	for (i = 0; i < 9; i++) {
		if (mask & 1) {
			redraw_tile(state->goalwin, i%3, i/3,
			            x+2, gp[0], gp[1], gp[2], 0);
		}
		mask >>= 1;
	}
}

static WINDOW *realloc_area(WINDOW **orig, int h, int w, int y, int x)
{
	WINDOW *win = *orig;

	if (win) {
#if HAVE_CURSES_WRESIZE
		if (wresize(win, h, w) != ERR) {
			mvwin(win, y, x);
			return win;
		}
#endif
		delwin(win);
	}

	if (w > 0 && h > 0)
		return *orig = subwin(stdscr, h, w, y, x);
	return *orig = NULL;
}

static void realloc_tiles(WINDOW **win, int h)
{
	WINDOW *border = win[WINDOW_TILEBORDER], *fill = win[WINDOW_TILEFILL];
	int w = 2*h - 1;

	if (fill && border) {
		int old_w, old_h;

#if HAVE_CURSES_WRESIZE
		if (wresize(fill, h-2, w-2) != ERR
		    && wresize(border, h, w) != ERR)
		{
			return;
		}
#endif

		getmaxyx(border, old_h, old_w);
		if (old_h == h)
			return;
	}

	if (fill)
		delwin(fill);
	if (border)
		delwin(border);

	win[WINDOW_TILEBORDER] = border = newwin(h, w, 0, 0);
	win[WINDOW_TILEFILL] = derwin(border, h-2, w-2, 1, 1);
}

static void setup_mainwin(struct app_state *state)
{
	int w, h, gamesz, goalsz, scr_w, scr_h, split;

	getmaxyx(stdscr, scr_h, scr_w);

	/* First try to fit the game tiles based on window height. */
	gamesz = MAX(3, (scr_h - 4) / 5);

	/* Adjust downward until we can fit smallest possible goal area. */
	for (; split = 5+10*gamesz, gamesz > 3; gamesz--) {
		if (split + 20 < scr_w)
			break;
	}

	/* Pick a goal size that will fit in the remaining area */
	goalsz = MAX(3, (scr_w - split - 4) / 6);
	if (goalsz >= gamesz)
		goalsz = MAX(3, gamesz - 1);

	realloc_tiles(state->gamewin, gamesz);
	realloc_tiles(state->goalwin, goalsz);

	/* Frame for game area */
	w = MIN(scr_w-2, 3+10*gamesz);
	h = MIN(scr_h-1, 2+5*gamesz);
	realloc_area(&state->gamewin[WINDOW_AREA], h, w, 1, 2);

	/* Frame for goal area */
	w = MIN(scr_w-split, 3+6*goalsz);
	h = MIN(scr_h-1, 2+3*goalsz);
	realloc_area(&state->goalwin[WINDOW_AREA], h, w, 1, split);
}

static void app_initialize(int argc, char **argv)
{
	int enable_mouse = 1;
	int opt;

	if (argc > 0)
		progname = argv[0];

	while ((opt = getopt_long(argc, argv, SOPT_STRING, lopts, 0)) != -1) {
		switch (opt) {
		case LOPT_MOUSE:
			enable_mouse = 2;
			break;
		case LOPT_NO_MOUSE:
			enable_mouse = 0;
			break;
		case LOPT_VERSION:
			print_version();
			exit(EXIT_SUCCESS);
		case LOPT_HELP:
			print_help();
			exit(EXIT_SUCCESS);
		default:
			print_usage(stderr);
			exit(EXIT_FAILURE);
		}
	}

	game_reset(&state.board);
	state.cursor = 5*state.board.y + state.board.x;

	initscr();
	start_color();
	curs_set(0);

	cbreak();
	keypad(stdscr, TRUE);
	if (enable_mouse) {
#if HAVE_CURSES_MOUSE_SET
		mouse_set(BUTTON1_PRESSED);
#elif HAVE_CURSES_MOUSEMASK
		mousemask(BUTTON1_PRESSED, NULL);
#endif
#if HAVE_CURSES_MOUSEINTERVAL
		mouseinterval(0);
#endif
	}

	noecho();

	init_pair(TILE_RED, COLOR_RED, COLOR_BLACK);
	init_pair(TILE_ORANGE, COLOR_YELLOW, COLOR_BLACK);
	init_pair(TILE_YELLOW, COLOR_YELLOW, COLOR_BLACK);
	init_pair(TILE_GREEN, COLOR_GREEN, COLOR_BLACK);
	init_pair(TILE_BLUE, COLOR_BLUE, COLOR_BLACK);
	init_pair(TILE_WHITE, COLOR_WHITE, COLOR_BLACK);
	init_pair(TILE_MAX, COLOR_BLACK, COLOR_BLACK);

	setup_mainwin(&state);
	refresh();
}

static uint_fast32_t do_move(struct app_state *state, int x, int y)
{
	uint_fast32_t mask;

	if ((mask = game_do_move(&state->board, x, y)) != 0) {
		curs_redraw_game(state, mask);
		refresh();
	}

	return mask;
}

#if HAVE_CURSES_MOUSE_SUPPORT
static void do_mouse(struct app_state *state)
{
	unsigned long bstate;
	int x, y;

#if HAVE_CURSES_GETMOUSE_NCURSES
	MEVENT mev;

	if (getmouse(&mev) == ERR)
		return;

	x = mev.x, y = mev.y;
	bstate = mev.bstate;
#elif HAVE_CURSES_REQUEST_MOUSE_POS
	request_mouse_pos();
	x = MOUSE_X_POS;
	y = MOUSE_Y_POS;

	bstate = 0;
	if (BUTTON_CHANGED(1)) {
		switch (BUTTON_STATUS(1)) {
		case BUTTON_RELEASED: bstate |= BUTTON1_RELEASED;
		case BUTTON_PRESSED:  bstate |= BUTTON1_PRESSED;
		}
	}
#endif
	if (bstate == BUTTON1_PRESSED) {
		uint_fast32_t cursor_mask, move_mask;
		int w, h;

		/* Determine size of the game area */
		getmaxyx(state->gamewin[WINDOW_TILEBORDER], h, w);
		w = 2*(w+1)/2;

		if (x < 4 || (x -= 4)/5 >= w) return;
		if (y < 2 || (y -= 2)/5 >= h) return;

		/* Turn off the keyboard cursor when using the mouse */
		cursor_mask = state->cursor < 0 ? -1 : 1ul << state->cursor;
		state->cursor = -1;

		move_mask = do_move(state, x/w, y/h);
		if ((cursor_mask & move_mask) == 0) {
			curs_redraw_game(state, cursor_mask);
			refresh();
		}
	}
}
#endif

static void do_move_cursor(struct app_state *state, int c)
{
	uint_fast32_t mask = 1ul << state->cursor;

	if (state->cursor < 0) {
		/* Reset keyboard cursor to the empty position */
		state->cursor = 5*state->board.y + state->board.x;
	}

	switch (c) {
	case KEY_UP:
		if ((state->cursor -= 5) < 0)
			state->cursor += 25;
		break;
	case KEY_DOWN:
		if ((state->cursor += 5) >= 25)
			state->cursor -= 25;
		break;
	case KEY_LEFT:
		if ((state->cursor -= 1) % 5 == 4 || state->cursor < 0)
			state->cursor += 5;
		break;
	case KEY_RIGHT:
		if ((state->cursor += 1) % 5 == 0)
			state->cursor -= 5;
		break;
	}

	curs_redraw_game(state, mask | 1ul << state->cursor);
	refresh();
}

static void do_keystroke(struct app_state *state, int c)
{
	switch (c) {
	case KEY_DOWN: case KEY_UP: case KEY_LEFT: case KEY_RIGHT:
		do_move_cursor(state, c);
		break;
	case ' ':
		if (state->cursor >= 0)
			do_move(state, state->cursor%5, state->cursor/5);
		break;
	}
}

int main(int argc, char **argv)
{
	setlocale(LC_ALL, "");
	app_initialize(argc, argv);

	curs_redraw_game(&state, -1);
	curs_redraw_goal(&state, -1);
	refresh();

	while (1) {
		int c = getch();

		switch (c) {
#ifdef KEY_RESIZE
		case KEY_RESIZE:
			setup_mainwin(&state);
			clear();
			refresh();
			curs_redraw_game(&state, -1);
			curs_redraw_goal(&state, -1);
			refresh();
			break;
#endif
#if HAVE_CURSES_MOUSE_SUPPORT
		case KEY_MOUSE:
			do_mouse(&state);
			break;
#endif
		default:
			do_keystroke(&state, c);
		}
	}
}
